import os
import sys
import re

from .base_package import Service as BasePackageService

class Service(BasePackageService):
    def __init__(self, config=None, logger=None):
        super().__init__(config=config, logger=logger)

        self.sbt_credentials_path = os.getenv('SBT_CREDENTIALS_PATH', '%s/.sbt/.credentials' % os.environ['HOME'])

        self.name_regex = re.compile('^\s*name\s*:=\s*[\'\"]+(.*?)[\'\"].*$', re.MULTILINE)
        self.version_regex = re.compile('^\s*version.* := [\'\"]+(\d+\.\d+\.\d+\.*\d*)(?:-SNAPSHOT)?[\'\"].*$', re.MULTILINE)
        self.scala_version_regex = re.compile('^\s*(?:val )?scala(?:Version)?\s+:?= [\'\"]+(\d+\.\d+\.\d+)[\'\"].*$', re.MULTILINE)
        self.scala_cross_version_regex = re.compile('^\s*crossScalaVersions\s+:=\s+Seq\((.*)\).*$', re.MULTILINE)

        self.package_version_format = self.config.get('package_version_format', 'version in ThisBuild := "%s"')

    def generate_sbt_credentials(self, credentials=None):
        if credentials:
            self.credentials = credentials
        else:   # setup defaults
            self.credentials = {
                'realm': os.environ['SBT_CREDENTIALS_REALM'],
                'host': os.environ['SBT_CREDENTIALS_HOST'],
                'user': os.environ['SBT_CREDENTIALS_USER'],
                'password': os.environ['SBT_CREDENTIALS_PASSWORD'],
            }

        self.logger.log('generating credentials at %s' % self.sbt_credentials_path)
        os.makedirs(os.path.dirname(self.sbt_credentials_path), exist_ok=True)
        with open(self.sbt_credentials_path, 'w') as creds:
            for key, value in self.credentials.items():
                creds.write('%s=%s\n' % (key, value))

        self.logger.log('credentials written to %s' % self.sbt_credentials_path)

    def read_package_info(self, package_info={}):
        package_file_paths = self.list_file_paths(self.config['package_directory'], '**/build.sbt')

        scala_versions = None

        if 'scala_version' in self.config['package_info']:
            scala_versions = self.config['package_info']['scala_version']

        scala_version_paths = [
            '%s/build.sbt' % self.config['package_directory'],
        ]
        for path in scala_version_paths:
            if scala_versions:
                break

            scala_versions = self.scala_versions(path)

        if not scala_versions:
            raise ValueError('Scala version not defined at: %s' % scala_version_paths)

        if not isinstance(scala_versions, list):
            scala_versions = [scala_versions]

        self.package_info = []
        for package_file_path in package_file_paths:
            directory = os.path.dirname(package_file_path)

            package_version_path = '%s/version.sbt' % directory
            if os.path.exists(package_version_path):
                for sv in scala_versions:
                    pi = self.config['package_info'].copy()
                    pi['name'] = self.package_name(package_file_path)
                    pi['scala_version'] = sv
                    pi['version'] = self.package_version(package_version_path)
                    pi['version_file_path'] = package_version_path
                    self.package_info.append(pi)
                    self.results[pi['name']] = pi

    def scala_versions(self, path):
        with open(path) as f:
            text = f.read()

            cross_version_result = self.scala_cross_version_regex.search(text)
            if cross_version_result:
                return ''.join(cross_version_result.groups()[0].split()).replace('"', '').split(',')

            result = self.scala_version_regex.search(text)
            if result:
                return result.groups()[0]

    # only works with Artifactory :(
    def get_existing_versions(self, **package_info):
        import requests
        from requests.auth import HTTPBasicAuth

        minor_scala_version = '.'.join(package_info['scala_version'].split('.')[0:-1])
        package_path = '%s/%s_%s' % (package_info['organization'], package_info['name'], minor_scala_version)
        url = '%s/api/storage/%s/%s' % (os.environ['ARTIFACTORY_ENDPOINT'], package_info['repository'], package_path)

        response = requests.get(
            url,
            auth=HTTPBasicAuth(os.environ['ARTIFACTORY_USERNAME'], os.environ['ARTIFACTORY_PASSWORD']),
        )

        versions = []
        if response.status_code == 404:
            self.logger.log('WARNING: repository path not found: %s' % url, level='warn')
        else:
            response.raise_for_status()

            body = response.json()
            versions.extend([child['uri'][1:] for child in body['children'] if child['folder']])

        return versions



from .base_auth_adapter import BaseAuthAdapter

import urllib.request
import os
from servicer.run import run

class AuthAdapter(BaseAuthAdapter):
    def __init__(self, config, logger=None):
        super().__init__(config, logger=logger)
        self.run = run

        self.version = self.config['version']
        self.architecture = self.config.get('architecture', 'amd64')
        self.os = self.run('uname -s')['stdout'].strip().lower()
        self.path = self.config.get('path', '/usr/local/bin/')
        self.name = 'packer_%s_%s_%s' % (self.version, self.os, self.architecture)

        if self.path.endswith('/'):
            self.path += self.name

    def authenticate(self):
        result = self.run('packer --version', check=False)
        if result['status'] == 0 and result['stdout'].strip() == self.version:
            self.logger.log('packer version %s is already installed' % self.version)
        else:
            self.install()

    def install(self):
        self.logger.log('installing packer version: %s' % self.version)

        download_file = '%s.zip' % self.name
        download_directory = os.path.dirname(self.path)
        download_path = '%s/%s' % (download_directory, download_file)
        url = 'https://releases.hashicorp.com/packer/%s/%s' % (self.version, download_file)

        self.run('rm -f %s/packer' % download_directory)
        file_path = '%s.zip' % self.path
        self.logger.log('downloading packer from %s to %s' % (url, download_path))
        urllib.request.urlretrieve(url, download_path)

        self.run('unzip -d `dirname %s` %s' % (download_path, download_path))
        self.run('rm %s' % download_path)



from .base_ci_adapter import BaseCIAdapter

class CIAdapter(BaseCIAdapter):

    def __init__(self, logger=None):
        super().__init__(logger=logger)

        # https://circleci.com/docs/2.0/env-vars/#built-in-environment-variables
        self.env_map['CIRCLE_BRANCH'] = 'BRANCH'
        self.env_map['CIRCLE_BUILD_NUM'] = 'BUILD_NUMBER'
        self.env_map['CIRCLE_BUILD_URL'] = 'BUILD_URL'
        self.env_map['CIRCLE_COMPARE_URL'] = 'DIFF_URL'
        self.env_map['CIRCLE_INTERNAL_TASK_DATA'] = 'RESULT_DATA'
        self.env_map['CIRCLE_JOB'] = 'JOB_NAME'
        self.env_map['CIRCLE_NODE_INDEX'] = 'NODE_INDEX'
        self.env_map['CIRCLE_NODE_TOTAL'] = 'NODE_TOTAL'
        self.env_map['CIRCLE_PR_NUMBER'] = 'PR_NUMBER'
        self.env_map['CIRCLE_PR_REPONAME'] = 'PR_REPO_NAME'
        self.env_map['CIRCLE_PR_USERNAME'] = 'PR_USERNAME'
        self.env_map['CIRCLE_PREVIOUS_BUILD_NUM'] = 'PREVIOUS_BUILD_NUMBER'
        self.env_map['CIRCLE_PROJECT_REPONAME'] = 'REPO_NAME'
        self.env_map['CIRCLE_PROJECT_USERNAME'] = 'CI_PROJECT_NAME'
        self.env_map['CIRCLE_PULL_REQUEST'] = 'PR_URL'
        self.env_map['CIRCLE_PULL_REQUESTS'] = 'PR_URLS'
        self.env_map['CIRCLE_REPOSITORY_URL'] = 'REPO_URL'
        self.env_map['CIRCLE_SHA1'] = 'COMMIT'
        self.env_map['CIRCLE_TAG'] = 'TAG'
        self.env_map['CIRCLE_USERNAME'] = 'USERNAME'
        self.env_map['CIRCLE_WORKFLOW_ID'] = 'WORKFLOW_ID'
        self.env_map['CIRCLE_WORKING_DIRECTORY'] = 'WORKING_DIRECTORY'



from ..service import Service as BaseService

class GCloudService(BaseService):
    def __init__(self, config, logger=None):
        super().__init__(config, logger=logger)
        self.name = 'a-gcloud-service'



class Logger():
    def __init__(self, level='info'):
        self.level = level

        self.levels = [
            'debug',
            'info',
            'warn',
            'error',
        ]

    def log(self, message='', level='info'):
        if self.levels.index(level) >= self.levels.index(self.level):
            print(message)



import requests
import json

from .service import Service as BaseService

class Service(BaseService):
    def request(self, **params):
        raise_for_status = params.pop('raise_for_status', True)
        method = params.pop('method', 'get')
        url = params.pop('url')

        if getattr(self, 'base_url'):
            url = '%s/%s' % (self.base_url, url)

        response = getattr(requests, method)(url, **params)
        if raise_for_status:
            response.raise_for_status()
        return response



import os

from servicer.run import run
from servicer.token_interpolator import TokenInterpolator

class Service:
    def __init__(self, config=None, logger=None):
        self.logger = logger
        self.run = run

        if config == None:
            return

        self.config = config

        self.project = os.environ['PROJECT_NAME']
        self.environment = os.getenv('SERVICE_ENVIRONMENT')

        self.token_interpolator = TokenInterpolator()

    def up(self):
        pass

    def down(self):
        pass



import os
import glob
import json
import re

from ..service import Service as BaseService
from servicer.git import Git

class Service(BaseService):
    def __init__(self, config=None, logger=None):
        super().__init__(config=config, logger=logger)
        self.package_version_format = None

        if 'git' in config:
            self.git = config['git']['module']

        if 'package_info' not in self.config:
            self.config['package_info'] = {}

        if 'package_file_path' in self.config['package_info'] and 'version_file_path' not in self.config['package_info']:
             self.config['package_info']['version_file_path'] = self.config['package_info']['package_file_path']

    def up(self):
        super().up()

        self.results = {}

        if 'steps' in self.config:
            self.run_steps(self.config['steps'])

        return self.results

    def run_steps(self, steps):
        for step in steps:
            getattr(self, step['type'])(**step.get('args', {}))

    def set_auto_version(self, max_increment=10, auto_detect_version=True):
        self.logger.log('auto-versioning (auto_detect_version=%s)...' % auto_detect_version)

        self.read_package_info(self.config['package_info'])
        if not isinstance(self.package_info, list):
            self.package_info = [self.package_info]

        if auto_detect_version:
            current_increment = 0
            version_changed = False
            while True:
                any_invalid_version = False
                for pi in self.package_info:

                    pi['version_exists'] = self.if_package_version_exists(**pi)
                    if pi['version_exists']:
                        any_invalid_version = True

                if not any_invalid_version:
                    break

                if current_increment > max_increment:
                    raise ValueError('Max package_version auto-increment reached! %s-%s' % (self.package_info[0]['package_name'], self.package_info[0]['version']))

                for pi in self.package_info:
                    if 'version_exists' in pi and pi['version_exists']:
                        pi['version'] = self.increment_version(pi['version'])

                version_changed = True
                current_increment += 1
        else:
            for pi in self.package_info:
                pi['version'] = self.increment_version(pi['version'])
            version_changed = True

        self.logger.log('Automatic version decided: %s' % ['%s:%s' % (pi['name'], pi['version']) for pi in self.package_info])

        if version_changed:
            if 'changed_files' not in self.results:
                self.results['changed_files'] = set()

            for pi in self.package_info:
                self.write_package_version(
                    path=pi['version_file_path'],
                    version=pi['version'],
                )

                self.results['changed_files'].add(pi['version_file_path'])

            self.results['changed_files'] = list(self.results['changed_files'])

    def commit_and_push_changes(self, git_no_verify=False, terminate_on_change=False):
        if 'BRANCH' not in os.environ:
            self.logger.log('No BRANCH defined, skipping commit and push.')
            return

        if self.git.protocol == 'https' and 'GIT_USERNAME' not in os.environ:
            self.logger.log('No GIT_USERNAME defined, skipping commit and push.')
            return

        commit_args = {
            'message': '[servicer] Automated version change.',
        }

        if 'version_file_path' in self.config['package_info']:
            commit_args['add'] = self.config['package_info']['version_file_path']

        commit_result = self.git.commit(**commit_args)
        push_result = self.git.push(ref=os.environ['BRANCH'], no_verify=git_no_verify)

        if commit_result['status'] == 0:
            if terminate_on_change:
                # requests termination of the build after the current step completes
                os.environ['TERMINATE_BUILD'] = '0'
        else:
            self.logger.log('no changes to commit')

    def increment_version(self, version):
        new_version = [int(v) for v in version.split('.')]
        new_version[-1] += 1
        return '.'.join([str(v) for v in new_version])

    def if_package_version_exists(self, **package_info):
        self.logger.log('checking for %s-%s...' % (package_info['name'], package_info['version']))

        version_list_method_name = "get_existing_versions"
        if 'existing_versions_source' in self.config:
            version_list_method_name = "get_existing_%s_versions" % self.config['existing_versions_source']

        version_list_method = getattr(self, version_list_method_name)
        version_list = version_list_method(**package_info)

        self.logger.log('existing versions: %s' % version_list)

        package_info['version_exists'] = package_info['version'] in version_list
        if package_info['version_exists'] and 'action' in package_info:
            if args['action'] == 'error':
                raise ValueError('Package already exists! %s-%s' % (package_info['name'], package_info['version']))

        return package_info['version_exists']

    def get_existing_gcr_versions(self, **package_info):
        docker_image = package_info['docker_image_path']
        if 'gcr_package_subdir' in package_info and package_info['gcr_package_subdir']:
            docker_image = '%s/%s' % (docker_image, package_info['name'])

        result = self.run('gcloud container images list-tags %s --format=json' % docker_image, hide_output=True)

        tags = set()
        for r in json.loads(result['stdout']):
            tags.update(r['tags'])

        return list(tags)

    def read_package_info(self, package_info={}):
        self.package_info = package_info

        if 'name' not in self.package_info:
            self.package_info['name'] = self.package_name(self.config['package_info']['package_file_path'])

        if 'package_version_format' in self.package_info:
            self.package_version_format = self.package_info['package_version_format']

        if not self.package_version_format:
            raise ValueError('Service must provide package_version_format.')

        if 'version' not in self.package_info:
            if 'version_regex' in self.package_info:
                self.version_regex = re.compile(self.package_info['version_regex'])
            self.package_info['version'] = self.package_version(self.config['package_info']['version_file_path'])
            self.logger.log('package version is %s' % self.package_info['version'])

        self.results['package_name'] = self.package_info['name']
        if isinstance(self.results['package_name'], list):
            self.results['package_name'] = ','.join(self.results['package_name'])
        self.results['package_version'] = self.package_info['version']

        pieces = self.package_info['version'].split('.')
        self.results['package_version_major'] = pieces[0]
        if len(pieces) > 1:
            self.results['package_version_minor'] = pieces[1]
        if len(pieces) > 2:
            self.results['package_version_patch'] = pieces[2]
        if len(pieces) > 3:
            self.results['package_version_revision'] = pieces[3]

    def package_name(self, path):
        with open(path) as f:
            text = f.read()
            result = self.name_regex.search(text)

            if result:
                name = result.groups()[0]
                return name
            else:
                raise ValueError('Package name not defined at: %s' % path)

    def package_version(self, path):
        with open(path) as f:
            text = f.read()
            result = self.version_regex.search(text)

            if result:
                version = result.groups()[0]
                return version
            else:
                raise ValueError('Package version not defined at: %s' % path)

    def write_package_version(self, path, version):
        text = ''
        with open(path) as f:
            text = f.read()

        new_version = self.package_version_format % version
        text = self.version_regex.sub(new_version, text)

        with open(path, 'w') as out:
            out.write(text)
            self.results['package_version'] = version

    def list_file_paths(self, path, match_glob):
        return glob.glob('%s/%s' % (path, match_glob), recursive=True)



import os
import sys

PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
print(SCRIPT_DIR)
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))
print(sys.path)

from .servicer import Servicer

def main():
    servicer = Servicer()
    servicer.run_steps()



from functools import reduce

# destructive topological dependency sort
# expects a dict containing dependency sets
# http://rosettacode.org/wiki/Topological_sort#Python
def toposort2(data):
    result = []

    if not data:
        return result

    for k, v in data.items():
        v.discard(k) # Ignore self dependencies

    # ensure all referenced items have an entry in the data dict
    extra_items_in_dependencies = reduce(set.union, data.values()) - set(data.keys())
    data.update({item: set() for item in extra_items_in_dependencies})

    # perform the topological ordering
    while True:
        ordered = set(item for item, dependency in data.items() if not dependency)
        if not ordered:
            break

        # build the next topological layer
        result.append(sorted(ordered))

        # remove the current layer's dependencies from each item
        data = {item: (dep - ordered) for item, dep in data.items() if item not in ordered}

    assert not data, 'A cyclic dependency exists amongst %r' % data

    return result

if __name__ == '__main__':
    # sample execution
    data = {
        'des_system_lib':   set('std synopsys std_cell_lib des_system_lib dw02 dw01 ramlib ieee'.split()),
        'dw01':             set('ieee dw01 dware gtech'.split()),
        'dw02':             set('ieee dw02 dware'.split()),
        'dw03':             set('std synopsys dware dw03 dw02 dw01 ieee gtech'.split()),
        'dw04':             set('dw04 ieee dw01 dware gtech'.split()),
        'dw05':             set('dw05 ieee dware'.split()),
        'dw06':             set('dw06 ieee dware'.split()),
        'dw07':             set('ieee dware'.split()),
        'dware':            set('ieee dware'.split()),
        'gtech':            set('ieee gtech'.split()),
        'ramlib':           set('std ieee'.split()),
        'std_cell_lib':     set('ieee std_cell_lib'.split()),
        'synopsys':         set(),
    }

    import pprint
    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(data)

    result = toposort2(data)
    print('result:')
    pp.pprint(result)



import glob
import os

from .service import Service as BaseService

class Service(BaseService):
    def up(self):
        super().up()

        if 'tasks' in self.config:
            self.run_tasks(self.config['tasks'])

    def run_tasks(self, tasks):
        for task in tasks:
            args = task.get('args', {})
            getattr(self, task['type'])(**args)

    def interpolate_file(self, files=[], ignore_missing_key=False):
        if not isinstance(files, list):
            files = [files]

        for f in files:
            if 'output' not in f:
                f['output'] = f['input']

            input_files = glob.glob(f['input'])
            for i in input_files:
                output_path = f['output']
                if output_path.endswith('/'):
                    output_path = '%s%s' % (output_path, os.path.basename(i))

                self.logger.log('interpolating tokens (%s -> %s)' % (i, output_path))

                data = None
                with open(i, 'r') as in_file:
                    data = in_file.read()

                os.makedirs(os.path.dirname(output_path), exist_ok=True)
                with open(output_path, 'w') as out_file:
                    params = {**os.environ, **self.full_config}
                    data = self.token_interpolator.replace_tokens(data, params=params, ignore_missing_key=ignore_missing_key)
                    out_file.write(data)



# from .gcloud_service import GCloudService
from ..task_service import Service as BaseService
import os
import requests
from google.cloud import storage

class Service(BaseService):
    def up(self):
        self.project = self.config.get('project', os.environ['GCLOUD_PROJECT'])
        self.gs = storage.Client(project=self.project)

        self.bucket_name = self.config['bucket']
        self.bucket = storage.Bucket(self.gs, name=self.bucket_name)
        self.ensure_bucket()

        super().up()

    def down(self):
        super().down()

    def ensure_bucket(self):
        self.logger.log('ensuring bucket exists: %s' % self.bucket_name)

        region = self.config.get('region', os.getenv('GCLOUD_REGION'))
        if region:
            self.bucket.location = region

        if not self.bucket.exists():
            self.bucket.create()

    def blob(self, blob_arg, operation, operation_arg=None):
        op_args = []
        if operation_arg:
            arg = open(operation_arg, 'wb') if operation == 'download_to_file' else operation_arg
            op_args.append(arg)

        self.logger.log('executing blob operation: (%s).%s(%s)' % (blob_arg, operation, operation_arg))

        blob = self.bucket.blob(blob_arg)
        getattr(blob, operation)(*op_args)

        if operation_arg and operation == 'download_to_file':
            op_args[0].close()



import os
import sys
import importlib
import pkg_resources
import json
import argparse
import yaml
import re
import copy
import imp
from datetime import datetime

from .config_loader import ConfigLoader
from .dependency_grapher import DependencyGrapher
from .git import Git
from .run import run
from .token_interpolator import TokenInterpolator
from .logger import Logger

class Servicer():
    def __init__(self, args=None, init=True):
        if not init:
            return

        self.datetime = datetime

        self.version = pkg_resources.get_distribution('servicer').version
        self.run = run

        if args == None:
            args = vars(self.load_arguments())

        logger_params = {
            'level': args['log_level']
        }
        if os.getenv('DEBUG'):
            logger_params['level'] = 'debug'
        self.logger = Logger(**logger_params)

        self.token_interpolator = TokenInterpolator(logger=self.logger)

        if 'version' in args and args['version']:
            self.logger.log(self.version)
            sys.exit(0)

        self.logger.log('servicer version: %s' % self.version)

        self.load_environment(args)

        self.config_loader = ConfigLoader(args, logger=self.logger)
        self.config = self.config_loader.load_config()
        self.active_services = None

        self.normalize_ci_environment()
        self.determine_service_environment()

        self.config_loader.interpolate_config(self.config)

        self.logger.log('Services Config:', level='debug')
        self.logger.log(json.dumps(self.config, indent=4, sort_keys=True, default=str), level='debug')

        if 'show_config' in self.config['args'] and self.config['args']['show_config']:
            self.logger.log('Services Config:')
            self.logger.log(json.dumps(self.config, indent=4, sort_keys=True, default=str))
            sys.exit(0)

        self.git_init()

        self.decide_service_step_order()

    def decide_service_step_order(self):
        self.load_steps()

        if not self.active_services:
            self.active_services = self.load_service_modules()

        self.dependency_grapher = DependencyGrapher(
            self.config,
            self.active_services,
            self.steps,
            self.step_order,
            self.active_steps,
            logger=self.logger,
        )
        self.service_step_order = self.dependency_grapher.order_service_steps(self.active_services)

    def load_arguments(self):
        parser = argparse.ArgumentParser(description='Process deployment options.')

        parser.add_argument('-g', '--generate_ci', action='store_true', help='generate a ci config file, do not run any deploy options')
        parser.add_argument('-s', '--service', help='deploy only the provided service')
        parser.add_argument('-f', '--services_file', default='services.yaml', help='custom path to your services config file (default is services.yaml)')
        parser.add_argument('-p', '--servicer_config_path', default='%s/.servicer' % os.getcwd(), help='path to your servicer directory (default is ./.servicer)')
        parser.add_argument('--env_file_paths', default='%s/.servicer/.env.yaml:%s/.servicer/.env.yaml' % (os.getenv('HOME'), os.getcwd()), help='paths to your local .env files, colon-separated')
        parser.add_argument('--step', help='perform the comma-separated build steps, defaults to all steps')
        parser.add_argument('-c', '--show_config', action='store_true', help='prints the interpolated config file')
        parser.add_argument('-u', '--no_ignore_unchanged', '--no_cd', action='store_true', help='disables ignoring services through change detection')
        parser.add_argument('--no_tag', action='store_true', help='disables build tagging')
        parser.add_argument('--no_auth', action='store_true', help='disables build authentication, useful if you are already authenticated locally')
        parser.add_argument('-d', '--ignore_dependencies', action='store_true', help='disables automatic dependency execution')
        parser.add_argument('--tag', action='store_true', help='generate a git tag')
        parser.add_argument('-v', '--version', action='store_true', help='display the package version')
        parser.add_argument('-x', '--destroy', action='store_true', help='destroy the current service environment')
        parser.add_argument('--dry', action='store_true', help='skip all service-step actions')
        parser.add_argument('--log_level', default='info', help='set the desired logging level, options are: [info,debug,warn,error]')

        return parser.parse_args()

    def load_environment(self, args):
        os.environ['PROJECT_PATH'] = os.environ['PWD']
        os.environ['BUILD_DATETIME'] = str(self.datetime.utcnow())
        os.environ['BUILD_DATE'] = self.datetime.now().strftime('%Y-%m-%d')

        if 'env_file_paths' not in args:
            return

        for path in args['env_file_paths'].split(':'):
            self.load_env_file(path)

    def load_env_file(self, path):
        self.logger.log('checking for (.env.yaml) at (%s)' % path)
        if os.path.exists(path):
            self.logger.log('(.env.yaml) found, including these arguments:')

            yaml_dict = yaml.load(open(path))
            for key, value in yaml_dict.items():
                os.environ[key] = value
                self.logger.log(key)
            self.logger.log()

    def normalize_ci_environment(self):
        if 'ci' not in self.config:
            return

        self.print_title('normalizing CI environment')

        self.config['ci']['adapters'] = {}
        for p in self.config['ci']['providers']:
            self.logger.log('CI Adapter: %s' % p)
            ci_adapter_modules = [
                {
                    'name': 'ci_adapters.%s' % p,
                    'package': 'ci_adapters',
                    'file_path': '%s/ci_adapters/%s.py' % (self.config['config_path'], p),
                },
                {
                    'name': 'servicer.builtin.ci_adapters.%s' % p,
                    'package': 'servicer.builtin.ci_adapters',
                    'file_path': '%s/builtin/ci_adapters/%s.py' % (self.config['module_path'], p),
                },
            ]
            module = self.load_module_from_paths(ci_adapter_modules)
            self.config['ci']['adapters'][p] = module.CIAdapter(logger=self.logger)

        if 'generate_ci' in self.config['args'] and self.config['args']['generate_ci']:
            for ci_adapter in self.config['ci']['adapters'].values():
                self.active_services = self.config['services'].keys()
                self.decide_service_step_order()

                ci_adapter.generate_ci_config(self.config, self.service_step_order)
            sys.exit(0)

        for ci_adapter in self.config['ci']['adapters'].values():
            ci_adapter.convert_environment_variables()

    def determine_service_environment(self):
        self.logger.log()
        self.service_environment = os.getenv('SERVICE_ENVIRONMENT') or self.get_service_environment(os.getenv('BRANCH', 'local'))

        self.logger.log('branch: %s' % os.getenv('BRANCH'))
        self.logger.log('service environment: %s' % self.service_environment)
        if self.service_environment:
            os.environ['SERVICE_ENVIRONMENT'] = self.service_environment

    def get_service_environment(self, branch):
        service_environment = None
        if 'environment' not in self.config or 'mappings' not in self.config['environment']:
            return service_environment

        mappings = self.config['environment']['mappings']
        service_environment, self.service_environment_config = self.map_service_environment(branch, mappings)

        if service_environment:
            formatter = self.config['environment'].get('formatter')
            if formatter:
                if 'replace' in formatter:
                    for replacer in formatter['replace']:
                        for f in replacer['from']:
                            if f in service_environment:
                                service_environment = service_environment.replace(f, replacer['to'])

                if formatter.get('uppercase'):
                    service_environment = service_environment.upper()
                if formatter.get('lowercase'):
                    service_environment = service_environment.lower()

                truncate = formatter.get('truncate')
                if truncate:
                    service_environment = service_environment[:truncate]

            if 'variables' in self.service_environment_config:
                self.config_loader.load_environment_variables(self.service_environment_config['variables'])

        return service_environment

    def map_service_environment(self, branch, mappings=[]):
        for m in mappings:
            if 'branch' in m:
                if m['branch'].startswith('/') and m['branch'].endswith('/'):
                    regex = m['branch'][1:-1]
                else:
                    regex = '^%s$' % m['branch'].replace('*', '.*')

                result = re.match(regex, branch)
                if result:
                    return m.get('environment', branch), m
            elif 'tag' in m:
                # TODO: support tag mapping
                pass

        return None, None

    def git_init(self):
        if not 'git' in self.config or not self.config['git']['enabled']:
            return

        self.print_title('initializing git integration')

        git_args = { 'hide_output': 'DEBUG' not in os.environ, 'logger': self.logger }
        if 'protocol' in self.config['git']:
            git_args['protocol'] = self.config['git']['protocol']
        self.git = Git(**git_args)
        self.git.config = self.config['git']

        if 'auto-set-branch' in self.config['git'] and self.config['git']['auto-set-branch']:
            if 'BRANCH' not in os.environ:
                os.environ['BRANCH'] = self.git.current_branch()

        if 'auto-set-commit' in self.config['git'] and self.config['git']['auto-set-commit']:
            if 'COMMIT_SHORT' not in os.environ:
                os.environ['COMMIT_SHORT'] = self.git.current_commit(min_length=self.config['git']['commit-min-length'])
            if 'COMMIT_LONG' not in os.environ:
                os.environ['COMMIT_LONG'] = self.git.current_commit()
            if 'COMMIT' not in os.environ:
                os.environ['COMMIT'] = os.environ['COMMIT_SHORT']

        if 'config' in self.config['git']:
            for key, value in self.config['git']['config'].items():
                result = self.run('git config %s' % key, check=False)['stdout'].strip()
                if result == '':
                    self.run('git config %s "%s"' % (key, value))

        if 'fetch-all' in self.config['git'] and self.config['git']['fetch-all']:
            self.run('git fetch')
        elif 'fetch-tags' in self.config['git'] and self.config['git']['fetch-tags']:
            self.run('git fetch --tags')

        if 'tag' in self.config['args'] and self.config['args']['tag']:
            self.tag_build(check_git=False)
            sys.exit(0)

        if 'GIT_DIFF_REF' in os.environ:
            result = self.run('git cat-file -t %s' % os.environ['GIT_DIFF_REF'], check=False)
            if result['status'] != 0:
                self.logger.log('Invalid GIT_DIFF_REF provided!')
            else:
                self.config['git']['diff-ref'] = os.environ['GIT_DIFF_REF']

        if 'BRANCH' in os.environ:
            if self.config['git']['diff-tagging-enabled'] and 'diff-ref' not in self.config['git']:
                servicer_tag_part = 'servicer-%s' % self.git.sanitize_tag(os.environ['BRANCH'])
                self.build_tags = [t for t in self.git.list_tags() if t.startswith(servicer_tag_part)]

                self.logger.log('branch tag: %s' % servicer_tag_part, level='debug')
                self.logger.log('matching tags:', level='debug')
                self.logger.log('\n'.join(self.build_tags), level='debug')

                if len(self.build_tags) > 0:
                    self.config['git']['diff-ref'] = self.build_tags[-1]

            if 'diff-defaults-to-latest-tag' in self.config['git'] and self.config['git']['diff-defaults-to-latest-tag'] and 'diff-ref' not in self.config['git']:
                result = self.run('git describe --tags --abbrev=0 --match "servicer-*" HEAD', check=False)

                if result['status'] == 0:
                    latest_tag = result['stdout'].strip()
                    if latest_tag:
                        self.logger.log('defaulting to latest servicer git tag')
                        self.config['git']['diff-ref'] = latest_tag

            # TODO: remove this feature in next breaking update
            if 'default-branch' in self.config['git'] and self.config['git']['default-branch'] and 'diff-ref' not in self.config['git']:
                if os.environ['BRANCH'] != self.config['git']['default-branch']:
                    self.logger.log('defaulting Git Diff Ref to default-branch')
                    self.config['git']['diff-ref'] = 'origin/%s' % self.config['git']['default-branch']

        if 'diff-ref' in self.config['git']:
            self.logger.log('Git Diff Ref: %s\n' % self.config['git']['diff-ref'])

        if self.config['git']['ignore-servicer-commits'] and 'diff-ref' in self.config['git']:
            authors = self.git.authors_for_changes_ahead_of_ref(self.config['git']['diff-ref'])
            self.logger.log('Commit authors: %s' % authors)
            if 'servicer' in authors:
                self.logger.log('Automated servicer changes were detected, skipping this build.')
                sys.exit(0)

    def tag_build(self, check_git=True):
        if check_git:
            if not 'git' in self.config or not self.config['git']['enabled']:
                return

            if not self.config['git']['auto-tag'] or 'BUILD_NUMBER' not in os.environ:
                return

            if 'no_tag' in self.config['args'] and self.config['args']['no_tag']:
                return

        self.remove_stale_tags()
        servicer_tag = self.servicer_git_tag()

        if servicer_tag:
            self.logger.log('Tagging: %s' % servicer_tag)
            self.git.tag(servicer_tag, push=True)

    def remove_stale_tags(self):
        self.logger.log('Removing old tags...')

        build_tags = [t for t in self.git.list_tags() if t.startswith('servicer-')]
        self.logger.log('\nexisting servicer tags:', level='debug')
        self.logger.log('\n'.join(build_tags), level='debug')

        branches = ['/'.join(b.split('/')[1:]) for b in self.git.list_remote_branches()]
        tag_prefixes = [self.git.sanitize_tag(b) for b in branches]
        self.logger.log('\nexisting branch tag prefixes:', level='debug')
        self.logger.log('\n'.join(tag_prefixes), level='debug')

        valid_tags = set()
        for tp in tag_prefixes:
            prefix = 'servicer-%s' % tp
            for bt in build_tags:
                if bt.startswith(prefix):
                    valid_tags.add(bt)

        tags_to_delete = list(set(build_tags) - valid_tags)
        self.logger.log('\nstale tags for other branches:', level='debug')
        self.logger.log('\n'.join(tags_to_delete), level='debug')

        self.git.delete_tag(tags_to_delete)

        tag_prefix = self.git.sanitize_tag(os.environ['BRANCH'])
        build_tags_for_branch = [bt for bt in build_tags if bt.startswith('servicer-%s' % tag_prefix)]

        self.logger.log('\nstale tags for this branch: %s' % tag_prefix, level='debug')
        self.logger.log('\n'.join(build_tags_for_branch), level='debug')
        self.git.delete_tag(build_tags_for_branch)

    def servicer_git_tag(self):
        if 'BRANCH' not in os.environ:
            return

        sanitized_tag = self.git.sanitize_tag(os.environ['BRANCH'])
        return 'servicer-%s-%s-%s' % (sanitized_tag, os.environ['BUILD_DATE'], os.environ['BUILD_NUMBER'])

    def load_service_modules(self):
        self.print_title('loading service modules')

        if 'services' in self.config:
            for name, service in self.config['services'].items():
                service['name'] = name

        active_services = []
        if 'service' in self.config['args'] and self.config['args']['service']:
            active_services.extend(self.config['args']['service'].split(','))
        elif 'services' in self.config:
            active_services.extend(self.config['services'].keys())

        self.logger.log('Active Services:\n%s\n' % '\n'.join(active_services))

        self.ignore_unchanged_services(active_services)

        for service_name in active_services:
            service = self.config['services'][service_name]
            self.load_service_module(service)

        return active_services

    def load_service_module(self, service):
        service['module'] = None

        if 'config' not in service:
            service['config'] = {}

        if 'service_type' not in service:
            service['service_type'] = 'service'

        adapter_path = service['service_type']

        if 'provider' in service:
            adapter_path = '%s/%s' % (service['provider'], adapter_path)

            self.try_initialize_provider(service['provider'], service)

        if 'providers' in service:
            for provider in service['providers']:
                self.try_initialize_provider(provider, service)

        adapter_name = adapter_path.replace('/', '.')
        service_modules = [
            {
                'name': 'service_adapters.%s' % adapter_name,
                'package': 'service_adapters',
                'file_path': '%s/service_adapters/%s.py' % (self.config['config_path'], adapter_path),
            },
            {
                'file_path': '%s/service_adapters/%s.sh' % (self.config['config_path'], adapter_path),
            },
            {
                'name': 'servicer.builtin.service_adapters.%s' % adapter_name,
                'package': 'servicer.builtin.service_adapters',
                'file_path': '%s/builtin/service_adapters/%s.py' % (self.config['module_path'], adapter_path),
            },
            {
                'file_path': '%s/builtin/service_adapters/%s.sh' % (self.config['module_path'], adapter_path),
            },
        ]
        module = self.load_module_from_paths(service_modules)
        if isinstance(module, str):
            service['shell_script'] = module
        else:
            service['module'] = module

    def ignore_unchanged_services(self, services):
        if 'no_ignore_unchanged' in self.config['args'] and self.config['args']['no_ignore_unchanged']:
            return

        if not 'git' in self.config or not self.config['git']['enabled'] or not self.config['git']['ignore-unchanged']:
            return

        if 'diff-ref' not in self.config['git']:
            self.logger.log('No GIT_DIFF_REF found, aborting change detection.')
            return

        diff_files = self.git.diff(self.config['git']['diff-ref'], name_only=True, merge_base=True)
        self.logger.log('\nChanged Files:')
        self.logger.log('\n'.join(diff_files))

        # TODO: think through what top level 'watch_paths' means
        if 'ignore_paths' in self.config['git']:
            regexes = [self.sanitize_regex(matcher) for matcher in self.config['git']['ignore_paths']]
            matched_files, diff_files = self.match_regexes(diff_files, regexes)

        ignored_services = []
        for service_name in services:
            service = self.config['services'][service_name]
            service_changed_files = diff_files

            if 'git' in service:
                if 'watch_paths' in service['git']:
                    watch_regexes = [self.sanitize_regex(matcher) for matcher in service['git']['watch_paths']]

                    self.logger.log('\nService: %s' % service_name, level='debug')
                    self.logger.log('Matchers:', level='debug')

                    service_changed_files, _ = self.match_regexes(diff_files, watch_regexes)

                if 'ignore_paths' in service['git']:
                    ignore_regexes = [self.sanitize_regex(matcher) for matcher in service['git']['ignore_paths']]
                    _, service_changed_files = self.match_regexes(service_changed_files, ignore_regexes)

                if len(service_changed_files) > 0:
                    self.logger.log('\nChanged Files:', level='debug')
                    self.logger.log('\n'.join(service_changed_files), level='debug')
                else:
                    ignored_services.append(service_name)

        self.logger.log('\nIgnored Services:')
        for sn in ignored_services:
            self.logger.log(sn)
            services.remove(sn)

        self.logger.log('\nChanged Services:')
        self.logger.log('\n'.join(services))

    def sanitize_regex(self, matcher):
        if matcher.startswith('/') and matcher.endswith('/'):
            return matcher[1:-1]

        return matcher.replace('*', '.*')

    # takes a list of strings, and returns a tuple of strings that match and do not match
    def match_regexes(self, strings, regexes):
        if not isinstance(regexes, list):
            regexes = [regexes]

        matches = []
        unmatches = []
        for s in strings:
            for regex in regexes:
                result = re.match(regex, s)

                self.logger.log('%s ~= %s -> %s' % (regex, s, result), level='debug')

                if result:
                    matches.append(s)
                else:
                    unmatches.append(s)

        return matches, unmatches

    def try_initialize_provider(self, provider_name, service):
        if 'no_auth' in self.config['args'] and self.config['args']['no_auth']:
            self.logger.log('skipping provider initialization for %s (no_auth enabled, assuming already authenticated)' % provider_name)
            return

        if 'providers' not in self.config:
            return

        provider = self.config['providers'].get(provider_name)

        if provider == None:
            self.logger.log('ERROR: no provider definition found for: %s' % provider_name, level='error')
            return  # TODO: replace this return with sys.exit(1)

        provider['name'] = provider_name

        should_initialize = not provider.get('initialized')
        force_authenticate = False

        provider_config = copy.deepcopy(provider)

        if service and 'auth' in service:
            self.config_loader.merge_config(provider_config, service['auth'])
            should_initialize = True
            force_authenticate = True

        if should_initialize:
            self.initialize_provider(provider_config, force_authenticate=force_authenticate)
            provider['initialized'] = True

        if service:
            service['config']['initialized_provider'] = provider_config

    def initialize_provider(self, provider, force_authenticate=False):
        self.print_title('intializing provider: %s' % provider['name'])
        self.logger.log('force_authenticate: %s' % force_authenticate, level='debug')

        if 'libraries' in provider and provider['libraries']:
            self.run('%s install %s' % (os.getenv('PIP_EXE', 'pip'), ' '.join(provider['libraries'])))
        if 'auth_script' in provider and provider['auth_script']:
            self.logger.log('authenticating with: %s' % provider['name'])
            auth_script_paths = [
                { 'file_path': '%s/auth/%s.sh' % (self.config['config_path'], provider['name']) },
                { 'file_path': '%s/builtin/auth/%s.sh' % (self.config['module_path'], provider['name']) },
            ]
            auth_script_path = self.load_module_from_paths(auth_script_paths)
            provider['auth_script_path'] = auth_script_path
            self.run(auth_script_path)
        if 'config' in provider:
            auth_modules = [
                {
                    'name': 'auth_adapters.%s' % provider['name'],
                    'package': 'auth_adapters',
                    'file_path': '%s/auth_adapters/%s.py' % (self.config['config_path'], provider['name']),
                },
                {
                    'name': 'servicer.builtin.auth_adapters.%s' % provider['name'],
                    'package': 'servicer.builtin.auth_adapters',
                    'file_path': '%s/builtin/auth_adapters/%s.py' % (self.config['module_path'], provider['name']),
                },
            ]
            module = self.load_module_from_paths(auth_modules)
            auth = module.AuthAdapter(provider['config'], logger=self.logger)

            if force_authenticate or not auth.current_user():
                auth.authenticate()

            self.logger.log('Current User: %s' % auth.current_user())

    def load_module_from_paths(self, modules):
        for mp in modules:
            self.logger.log('searching for module at: %s' % mp['file_path'], level='debug')

            if os.path.exists(mp['file_path']):
                if 'name' in mp:
                    self.logger.log('importing: %s:%s' % (mp['name'], mp['package']), level='debug')

                    module = None
                    try:
                        module = importlib.import_module(mp['name'])
                    except ModuleNotFoundError:
                        self.logger.log('falling back to import via imp: %s' % mp['file_path'], level='debug')

                        with open(mp['file_path'], 'rb') as fp:
                            config_dir = self.config['args']['servicer_config_path'].split('/')[-1]
                            module = imp.load_module('%s/%s' % (config_dir, mp['name']), fp, mp['file_path'], ('.py', 'rb', imp.PY_SOURCE))

                    return module
                else:
                    self.logger.log('found matching executable: %s' % mp['file_path'])
                    return mp['file_path']

        self.logger.log('no modules found!')
        sys.exit(1)

    def load_steps(self):
        self.steps = {}
        self.step_order = []

        if 'steps' in self.config:
            for step in self.config['steps']:
                if step['name'] == 'destroy' and not ('destroy' in self.config['args'] and self.config['args']['destroy']):
                    continue

                self.steps[step['name']] = step
                self.step_order.append(step['name'])

        if 'step' in self.config['args'] and self.config['args']['step']:
            self.active_steps = self.config['args']['step'].split(',')
        else:
            self.active_steps = self.step_order.copy()

        print(self.active_steps)

    def run_service_steps(self):
        self.print_title('executing service-steps')
        self.logger.log(json.dumps(self.service_step_order, indent=4))
        self.logger.log()

        service_steps = [item for sublist in self.service_step_order for item in sublist]

        if 'destroy' in self.config['args'] and self.config['args']['destroy']:
            self.print_title('destroying services (in reverse dependency order)')
            destroyed_services = set()
            destroy_service_steps = []

            for ss in reversed(service_steps):
                service_name = ss.split(':')[0]
                if service_name not in destroyed_services:
                    destroyed_services.add(service_name)

                    if 'steps' in self.config['services'][service_name] and 'destroy' in self.config['services'][service_name]['steps']:
                        destroy_step = '%s:%s' % (service_name, 'destroy')
                        destroy_service_steps.append(destroy_step)

            service_steps = destroy_service_steps
            self.logger.log(json.dumps(service_steps, indent=4))

        for service_step_name in service_steps:
            ss_pieces = service_step_name.split(':')
            service_name = ss_pieces[0]
            step_name = ss_pieces[1]

            self.print_title('service-step: %s:%s' % (service_name, step_name))

            step = self.steps[step_name]
            service = self.config['services'][service_name]
            service_step = service['steps'][step_name]

            self.logger.log('step:', level='debug')
            self.logger.log(json.dumps(step, indent=4, sort_keys=True, default=str), level='debug')
            self.logger.log(level='debug')
            self.logger.log('service:', level='debug')
            self.logger.log(json.dumps(service, indent=4, sort_keys=True, default=str), level='debug')
            self.logger.log(level='debug')

            if 'config' in step and 'requires_service_environment' in step['config']:
                if step['config']['requires_service_environment'] and self.service_environment == None:
                    self.logger.log('skipping, no valid service environment found for step: %s' % step_name)
                    continue

            self.run_service_step(service, service_step)

            # TODO: rethink and standardize this termination process
            if 'TERMINATE_BUILD' in os.environ:
                self.logger.log('build termination requested, stopping with code: %s' % os.environ['TERMINATE_BUILD'])
                sys.exit(int(os.environ['TERMINATE_BUILD']))

        self.tag_build()
        self.logger.log('\nBuild Complete.')

    def run_service_step(self, service, service_step):
        if 'module' not in service:
            self.load_service_module(service)

        self.run_commands(service_step.get('commands'))

        if 'config' in service_step:
            config = service_step.get('config')

            if 'git' in self.config and self.config['git']['enabled']:
                if 'git' not in config:
                    config['git'] = {}
                config['git']['module'] = self.git

            interpolation_params = {**self.config, **os.environ}
            self.token_interpolator.interpolate_tokens(config, interpolation_params)

            adapter = service['module'].Service(config, logger=self.logger)
            adapter.full_config = self.config

            if 'dry' in self.config['args'] and self.config['args']['dry']:
                self.logger.log('DRY: adapter.up(): %s' % service.get('service_type'))
            else:
                results = adapter.up()

                if results:
                    service_step['results'] = results
                    self.logger.log('results: ')
                    self.logger.log(json.dumps(results, indent=4, sort_keys=True, default=str))

        self.run_commands(service_step.get('post_commands'))

    def run_commands(self, commands):
        if not commands:
            return

        for c in commands:
            if isinstance(c, dict):
                if 'env_var' in c:
                    result = self.run_command(c['command'])
                    os.environ[c['env_var']] = result['stdout'].strip()
            else:
                self.run_command(c)

    def run_command(self, command):
        interpolation_params = {**self.config, **os.environ}
        command = self.token_interpolator.replace_tokens(command, interpolation_params)

        if 'dry' in self.config['args'] and self.config['args']['dry']:
            self.logger.log('DRY: executing: %s' % command)
            return

        return self.run(command)

    def print_title(self, message='', border='----'):
        inner_text = ' %s ' % message
        border_text = border

        while len(border_text) < len(inner_text):
            border_text = '%s %s' % (border_text, border)

        self.logger.log()
        self.logger.log(border_text)
        self.logger.log(inner_text)
        self.logger.log(border_text)
        self.logger.log()

def main():
    Servicer().run_service_steps()

if __name__ == '__main__':
    main()



import os

class BaseAuthAdapter():

    def __init__(self, config, logger=None):
        self.logger = logger
        self.config = config

    def authenticate(self):
        pass

    def current_user(self):
        pass



import textwrap

from .base_ci_adapter import BaseCIAdapter

class CIAdapter(BaseCIAdapter):

    def __init__(self, logger=None):
        super().__init__(logger=logger)

        # https://wiki.jenkins.io/display/JENKINS/Building+a+software+project
        self.env_map['BRANCH_NAME'] = 'BRANCH'
        self.env_map['BUILD_DISPLAY_NAME'] = 'BUILD_NAME'
        self.env_map['BUILD_ID'] = 'BUILD_ID'
        self.env_map['BUILD_NUMBER'] = 'BUILD_NUMBER'
        self.env_map['BUILD_TAG'] = 'TAG'
        self.env_map['BUILD_URL'] = 'BUILD_URL'
        self.env_map['CHANGE_AUTHOR'] = 'PR_USERNAME'
        self.env_map['CHANGE_AUTHOR_DISPLAY_NAME'] = 'PR_AUTHOR_NAME'
        self.env_map['CHANGE_AUTHOR_EMAIL'] = 'PR_AUTHOR_EMAIL'
        self.env_map['CHANGE_ID'] = 'PR_NUMBER'
        self.env_map['CHANGE_TARGET'] = 'PR_TARGET'
        self.env_map['CHANGE_TITLE'] = 'PR_TITLE'
        self.env_map['CHANGE_URL'] = 'PR_URL'
        self.env_map['EXECUTOR_NUMBER'] = 'NODE_INDEX'
        self.env_map['GIT_BRANCH'] = 'BRANCH'
        self.env_map['GIT_COMMIT'] = 'COMMIT'
        self.env_map['GIT_URL'] = 'REPO_URL'
        self.env_map['JENKINS_HOME'] = 'HOME'
        self.env_map['JENKINS_URL'] = 'CI_URL'
        self.env_map['JOB_BASE_NAME'] = 'JOB_BASE_NAME'
        self.env_map['JOB_NAME'] = 'JOB_NAME'
        self.env_map['JOB_URL'] = 'JOB_URL'
        self.env_map['NODE_LABELS'] = 'NODE_LABELS'
        self.env_map['NODE_NAME'] = 'NODE_NAME'
        self.env_map['WORKSPACE'] = 'WORKING_DIRECTORY'

    def generate_ci_config(self, config, service_step_order, path='./Jenkinsfile'):

        self.config = config
        self.indent_value = ' ' * self.config['ci'].get('indent_value', 2)

        servicer_body = [
            """checkout scm

servicer_image = '%s'
sh "docker pull $servicer_image"

""" % self.config['ci'].get('image', 'wmgroot/servicer-alpine'),
            {
                'block': [{
                    'function_name': [
                        'docker.image',
                        'inside',
                    ],
                    'params': [
                        'servicer_image',
                        "'-u root -v /var/run/docker.sock:/var/run/docker.sock -v /home/svc-jenkins/.ssh:/home/root/.ssh'",
                    ],
                    'body': self.servicer_commands(service_step_order),
                    'quote_strings': False,
                }],
            },
        ]

        if 'credentials' in self.config['ci']:
            wrap = '[\n%s%s\n]' % (self.indent_value, '%s')
            servicer_body = {
                'block': [{
                    'function_name': 'withCredentials',
                    'params': self.config['ci']['credentials'],
                    'wrap': wrap,
                    'body': servicer_body,
                }],
            }

        jenkins_body = {
            'block': [{
                'function_name': 'timeout',
                'params': {
                    'time': self.config['ci'].get('timeout', 2),
                    'unit': self.config['ci'].get('timeout_unit', 'HOURS'),
                },
                'body': {
                    'block': [{
                        'function_name': 'node',
                        'params': self.config['ci']['node'],
                        'body': {
                            'block': [{
                                'function_name': 'stage',
                                'params': 'Servicer',
                                'body': servicer_body,
                            }],
                        },
                    }],
                },
            }],
        }

        text = self.evaluate_body(jenkins_body)
        self.write_file(path, text)

    def servicer_commands(self, service_step_order):
        self.logger.log(service_step_order, level='debug')

        commands = ["\nsh 'gcloud auth activate-service-account --key-file=${GOOGLE_APPLICATION_CREDENTIALS}'"]

        for parallel_set in service_step_order:
            params = {}
            for service_step in parallel_set:
                pieces = service_step.split(':')
                command = "sh 'servicer --service=%s --step=%s --ignore_dependencies --no_tag'" % (pieces[0], pieces[1])
                params["'%s'" % service_step] = '{\n%s\n}' % textwrap.indent(self.block('stage', service_step, command), self.indent_value)

            commands.append('\n\n' + self.function_header(
                function_name='parallel',
                params=params,
                indent_params=True,
                quote_strings=False,
            ))

        return commands

    def evaluate_body(self, body):
        text = ''

        if isinstance(body, dict):
            for key, value in body.items():
                if hasattr(self, key):
                    for v in value:
                        if 'body' in v:
                            if not isinstance(v['body'], list):
                                v['body'] = [v['body']]

                            evaluated = ''

                            for b in v['body']:
                                if isinstance(b, dict):
                                    evaluated += self.evaluate_body(b)
                                else:
                                    evaluated += b

                            v['body'] = evaluated

                        result = getattr(self, key)(**v)
                        text += result

        return text

    def block(self, function_name, params={}, body='', wrap='%s', quote_strings=True):
        if not isinstance(function_name, list):
            function_name = [function_name]
            params = [params]

        i = 0
        headers = []
        while i < len(function_name):
            fn = function_name[0]
            headers.append(self.function_header(
                function_name[i],
                params[i],
                wrap=wrap,
                quote_strings=quote_strings,
            ))
            i += 1

        indented_body = textwrap.indent(body, self.indent_value)

        return """%s {
%s
}""" % ('.'.join(headers), indented_body)

    def function_header(
        self,
        function_name,
        params={},
        wrap='%s',
        joiner=', \n    ',
        indent_params=False,
        quote_strings=True,
    ):
        if not isinstance(params, list):
            params = [params]

        joined = []
        for p in params:
            if isinstance(p, dict):
                params_strings = []
                for key, value in p.items():
                    result = ''
                    if hasattr(self, key):
                        result = getattr(self, key)(**value)
                    elif isinstance(value, str) and quote_strings:
                        result = '%s: \'%s\'' % (key, value)
                    else:
                        result = '%s: %s' % (key, value)
                    params_strings.append(result)

                param_end = ',\n' if indent_params else ', '
                joined.append(param_end.join(params_strings))
            else:
                if not isinstance(params, list):
                    params = [params]
                params = ['\'%s\'' % p if isinstance(p, str) and quote_strings else p for p in params]
                joined.append(', '.join(params))

        wrapped = wrap % joiner.join(joined)

        if indent_params:
            wrapped = '\n%s\n' % textwrap.indent(wrapped, self.indent_value)

        return '%s(%s)' % (function_name, wrapped)

    def param_list(self, params={}, wrap='[%s]'):
        params_strings = []
        for key, value in params.items():
            result = ''
            if isinstance(value, str):
                result = '%s: \'%s\'' % (key, value)
            else:
                result = '%s: %s' % (key, value)
            params_strings.append(result)

        return wrap % ', '.join(params_strings)



from ..docker_image import Service as DockerImageService

class Service(DockerImageService):
    def __init__(self, config, logger=None):
        super().__init__(config, logger=logger)

        self.registry = 'us.gcr.io'

    def prune_images(self, older_than='', n_to_keep=0):
        full_path = self.full_image_path(self.config['registry_path'])

        command = "gcloud container images list-tags %s --limit=999999 --sort-by=TIMESTAMP --format='get(digest)'" % full_path

        if older_than:
            command += " --filter=\"timestamp.datetime < '%s'\"" % older_than

        result = self.run(command)
        shas = result['stdout'].strip().split('\n')

        n_to_delete = len(shas)
        if n_to_keep:
            count_result = self.run("gcloud container images list-tags %s --limit=999999 --format='get(digest)' | wc -l" % full_path)
            total_count = int(count_result['stdout'].strip())
            n_to_delete = total_count - n_to_keep

        if n_to_delete > 0:
            self.logger.log('pruning %s image digests from %s' % (n_to_delete, full_path))
            for digest in shas[:n_to_delete]:
                delete_command = 'gcloud container images delete -q --force-delete-tags "%s@%s"' % (full_path, digest)
                result = self.run(delete_command)
